//***************************************************************************
//
// BITSET.H
// by Harlen C. Batagelo
//
//***************************************************************************

#ifndef _BITSET_H
#define _BITSET_H

#include <math.h>

static size_t BITPOW[16] = 
{ 32768, 16384, 8192, 4096, 2048, 1024, 512, 256, 128, 64, 32, 16, 8, 4, 2, 1 };

static size_t BITMAX = 65535;

#define TYPE unsigned short int	// element type

/////////////////////////////////////////////////////////////////////////////
// Define a set of bits
/////////////////////////////////////////////////////////////////////////////
class BitSet
{
private:
	TYPE *bset;				// set of bits
	
	size_t n_bytes;			// number of allocated elements
	size_t t_size_b;		// size of TYPE
	size_t t_size;			// size of TYPE (in bits)
	size_t t_size_mo;		// size of TYPE (in bits) minus 1

public:
	BitSet (size_t n);
	~BitSet (void);

	// Return a pointer to the set of bits
	inline TYPE* getBitSet() 
	{ 
		return bset;
	}

	// Set the nth bit in the set
	inline void set (size_t n)
	{
		size_t offset = n / t_size; // element no.
		bset[offset] |= BITPOW[n - (offset * t_size)];
	}

	// Set ct bits starting from the nth bit
	inline void set (size_t n, size_t ct)
	{
		size_t ct_orig = ct;						// no. of elements to set
		size_t offset_start = n / t_size;			// starting element
		size_t offset_end = (n + ct - 1) / t_size;  // ending element
		size_t pos = n - (offset_start * t_size);	// bit pos at starting element

		// set first element
		for (size_t i = pos; i < t_size; i++, ct--)
		{
			if (i-pos == ct_orig) return;
			bset[offset_start] |= BITPOW[i];
		}

		// set middle
		size_t xx = ct / t_size;
		memset (bset+offset_start+1, BITMAX, t_size_b * xx);

		// set last element
		for	(i = 0; i < ct - (xx * t_size); i++)
			bset[offset_end] |= BITPOW[i];
	}

	// Reset ct bits starting from the nth bit
	inline void reset (size_t n, size_t ct)
	{
		size_t ct_orig = ct;						// no. of elements to reset
		size_t offset_start = n / t_size;			// starting element
		size_t offset_end = (n + ct - 1) / t_size;  // ending element
		size_t pos = n - (offset_start * t_size);	// bit pos at starting element

		// reset first element
		for (size_t i = pos; i < t_size; i++, ct--)
		{
			if (i-pos == ct_orig) return;
			bset[offset_start] &= BITPOW[BITMAX - i];
		}

		// reset middle
		size_t xx = ct / t_size;
		memset (bset+offset_start+1, 0, t_size_b * xx);

		// reset last element
		for	(i = 0; i < ct - (xx * t_size); i++)
			bset[offset_end] &= BITPOW[BITMAX - i];
	}


	// Reset the nth bit in the set
	inline void reset (size_t n)
	{
		size_t offset = n / t_size; // element no.
		bset[offset] &= BITPOW[BITMAX - (n - (offset * t_size))];
	}

	// Reset all bits in bit set
	inline void reset (void)
	{
		memset (bset, 0, t_size_b * n_bytes);
	}

	// Return value of the nth bit
	inline size_t operator[] (size_t n)
	{
		size_t offset = n / t_size;
		if (bset[offset] & BITPOW[n - (offset * t_size)])
			return 1;
		return 0;
	}

	// Copy memory between set of bits
	void getFromBitset (BitSet &bitset)
	{
		memcpy (bset, bitset.getBitSet(), t_size_b * n_bytes);
	}
};

#endif